%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% stereoOpt.m
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% CS223b
% Winter '05
%
% function [disparity_map, consistency, right_disp_map] =
% stereoOpt(left_im, right_im, maxdisp, window_size, check_consistency)
%
% Optimized dense stereo correlation function
%
% Dan Morris, Stanford University
%
% Computes a disparity map based on the stereo images 'left' and 'right',
% scanning for maxdisp pixels from each pixel in the left image.  If
% the original image is size (h,w), the disparity map will have size
% (h,w-maxdisp).
%
% The second return value is a consistency-checked version of the
% disparity map, in which all disparities that were not identical in 
% the left-to-right-matched map and the right-to-left-matched map
% are set to 0.
%
% window_size is the edge size of the (two-dimensional) neighbor-averaging
% filter that will be applied to the difference maps between the two images
% before computing the disparity maps.
%
% maxdisp defaults to 15, window_size defaults to 5
%
% The third matrix returned, right_disp_map, is the right-to-left-matched
% disparity map used in consistency checking
%
% If 'check_consistency' is zero, no consistency check is performed and the
% second and third return values are not computed.  'check_consistency'
% defaults to zero.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [disparity_map, consistency, right_disp_map] = ...
  stereoOpt(left_im, right_im, maxdisp, window_size, check_consistency)

% Check arguments and create defaults where we need to...
if (nargin < 2)
    fprintf(1,'correlation_match requires at least two arguments...\n');
    return;
end

if (nargin < 3)
    maxdisp = 15;
end

if (nargin < 4)
    window_size = 5;
end

if (nargin < 5)
  check_consistency = 0;
end

% Make sure the input images are the same size
eqim = (size(left_im)==size(right_im));
if (find(eqim==0))
    fprintf(1,'Images are not the same size...\n');
    return;
end

% Optionally apply a gaussian smoothing filter to the input images
SMOOTH_SIGMA = 0;

if (SMOOTH_SIGMA > 0) 
    ker = gauss(SMOOTH_SIGMA)' * gauss(SMOOTH_SIGMA);
    left_im = conv2(left_im,ker,'same');
    right_im = conv2(right_im,ker,'same');
end

% Create a matrix of size [height,width,maxdisp] to cache all the squared differences
dcache = zeros(size(left_im,1),size(left_im,2)-maxdisp,maxdisp+1);
if (check_consistency)
  % A second difference cache for the reverse disparity map
  dcache_r = dcache;
end

% Compute all left-to-right and right-to-left squared differences...
%
% I reverse the indexing in the right-to-left matched computation so the indices
% will have the same intuitive meaning as those used for the left-to-right matched
% differences (higher index into the resulting matrix is a higher disparity)
for(d=0:maxdisp)
    dcache(:,:,d+1) = (left_im(:,d+1:size(left_im,2)-maxdisp+d) - ...
      right_im(:,1:(size(left_im,2)-maxdisp))).^2;
    if (check_consistency)
      dcache_r(:,:,maxdisp+1-d) = (right_im(:,d+1:size(left_im,2)-maxdisp+d) - ...
        left_im(:,1+maxdisp:(size(left_im,2)))).^2;
    end       
end

% Convolve each isodisparity plane with an averaging filter (this is the
% windowing operation)
avg_filter = ones(window_size,window_size);
for(d=1:maxdisp+1)
    dcache(:,:,d) = conv2(dcache(:,:,d),avg_filter,'same');
    if (check_consistency)
      dcache_r(:,:,d) = conv2(dcache_r(:,:,d),avg_filter,'same');
    end
end

% Find the minimum difference at each x,y position; this is the disparity at
% that point
[diffs,inds] = min(dcache,[],3);
if (check_consistency)
  [diffs_r,inds_r] = min(dcache_r,[],3);
end

% Subtract 1 from each index, since the lowest possible disparity should be zero,
% not 1
inds = inds - 1;
if (check_consistency)
  inds_r = inds_r - 1;
end

% Do the consistency check if the caller asked us to...
if (check_consistency)
  
  % Compute a ramp vector for left-right consistency checking
  h = size(left_im,1);
  w = size(left_im,2) - maxdisp;
  ramp = linspace(1,h*w,h*w);
  ramp = reshape(ramp,h,w);

  % Evaluating the right disparity map at this vector _should+ yield the left
  % disparity map
  crosscheck = ramp - inds;
  crosscheck(find(crosscheck<1)) = 1;
  right_check = inds_r(crosscheck);
  
  % Anywhere we didn't find a match, we have a 'bad' disparity
  bad_inds = find(abs(right_check - inds) > 0);

  checked_left = inds;
  checked_left(bad_inds) = 0;

  % Set the output variables
  consistency = checked_left;
  right_disp_map = inds_r;
else
  consistency = 0;
  right_disp_map = 0;
end

% ...and return the disparity map...
disparity_map = inds;
